package ru.dbotthepony.kstarbound.io.json

import com.google.common.collect.ImmutableList
import com.google.common.collect.ImmutableSet
import com.google.gson.Gson
import com.google.gson.JsonArray
import com.google.gson.JsonElement
import com.google.gson.JsonNull
import com.google.gson.JsonObject
import com.google.gson.JsonPrimitive
import com.google.gson.JsonSyntaxException
import com.google.gson.TypeAdapter
import com.google.gson.TypeAdapterFactory
import com.google.gson.reflect.TypeToken
import com.google.gson.stream.JsonReader
import com.google.gson.stream.JsonToken
import com.google.gson.stream.JsonWriter
import it.unimi.dsi.fastutil.objects.ObjectOpenHashSet

fun <T> TypeAdapter<T>.transformRead(transformer: (T) -> T): TypeAdapter<T> {
	return object : TypeAdapter<T>() {
		override fun write(out: JsonWriter, value: T) {
			return this@transformRead.write(out, value)
		}

		override fun read(`in`: JsonReader): T {
			return transformer(this@transformRead.read(`in`))
		}
	}
}

fun <T> TypeAdapter<T>.transformWrite(transformer: (T) -> T): TypeAdapter<T> {
	return object : TypeAdapter<T>() {
		override fun write(out: JsonWriter, value: T) {
			return this@transformWrite.write(out, transformer(value))
		}

		override fun read(`in`: JsonReader): T {
			return this@transformWrite.read(`in`)
		}
	}
}

fun <In, Out> TypeAdapter<In>.transform(read: (In) -> Out, write: (Out) -> In): TypeAdapter<Out> {
	return object : TypeAdapter<Out>() {
		override fun write(out: JsonWriter, value: Out) {
			return this@transform.write(out, write(value))
		}

		override fun read(`in`: JsonReader): Out {
			return read(this@transform.read(`in`))
		}
	}
}

fun <T> TypeAdapter<T>.ifString(reader: (String) -> T): TypeAdapter<T> {
	return object : TypeAdapter<T>() {
		override fun write(out: JsonWriter, value: T) {
			return this@ifString.write(out, value)
		}

		override fun read(`in`: JsonReader): T {
			if (`in`.peek() == JsonToken.STRING) {
				return reader(`in`.nextString())
			}

			return this@ifString.read(`in`)
		}
	}
}

fun <T> TypeAdapter<T?>.neverNull(): TypeAdapter<T> {
	return object : TypeAdapter<T>() {
		override fun write(out: JsonWriter, value: T) {
			this@neverNull.write(out, value)
		}

		override fun read(`in`: JsonReader): T {
			val path = `in`.path
			return this@neverNull.read(`in`) ?: throw JsonSyntaxException("Value was null near $path")
		}
	}
}

fun <T> TypeAdapter<T>.allowNull(): TypeAdapter<T?> = nullSafe()

@Deprecated("Небезопасно, надо использовать другой путь")
fun TypeAdapterFactory.ifString(reader: (String) -> Any): TypeAdapterFactory {
	return object : TypeAdapterFactory {
		override fun <T : Any?> create(gson: Gson, type: TypeToken<T>): TypeAdapter<T>? {
			return this@ifString.create(gson, type)?.ifString(reader as (String) -> T)
		}
	}
}

fun JsonReader.consumeNull(): Boolean {
	if (peek() == JsonToken.NULL) {
		nextNull()
		return true
	}

	return false
}

fun JsonWriter.value(element: JsonPrimitive) {
	if (element.isBoolean) {
		value(element.asBoolean)
	} else if (element.isNumber) {
		value(element.asNumber)
	} else if (element.isString) {
		value(element.asString)
	} else {
		throw IllegalArgumentException(element.toString())
	}
}

fun JsonWriter.value(element: JsonNull) {
	nullValue()
}

fun JsonWriter.value(element: JsonArray) {
	beginArray()
	for (v in element) value(v)
	endArray()
}

fun JsonWriter.value(element: JsonObject) {
	beginObject()
	for ((k, v) in element.entrySet()) {
		name(k)
		value(v)
	}
	endObject()
}

fun JsonWriter.value(element: JsonElement?) {
	when (element) {
		is JsonPrimitive -> value(element)
		is JsonNull -> value(element)
		is JsonArray -> value(element)
		is JsonObject -> value(element)
		null -> nullValue()
		else -> throw IllegalArgumentException(element.toString())
	}
}

inline fun <reified C : Collection<E>, reified E> Gson.collectionAdapter(): TypeAdapter<C> {
	return getAdapter(TypeToken.getParameterized(C::class.java, E::class.java)) as TypeAdapter<C>
}

inline fun <reified E> Gson.listAdapter(): TypeAdapter<ImmutableList<E>> {
	return collectionAdapter()
}

inline fun <reified E> Gson.mutableListAdapter(): TypeAdapter<ArrayList<E>> {
	return collectionAdapter()
}

inline fun <reified E> Gson.setAdapter(): TypeAdapter<ImmutableSet<E>> {
	return collectionAdapter()
}

inline fun <reified E> Gson.mutableSetAdapter(): TypeAdapter<ObjectOpenHashSet<E>> {
	return collectionAdapter()
}

fun JsonArray(elements: Collection<JsonElement>): JsonArray {
	return JsonArray(elements.size).also { elements.forEach(it::add) }
}

fun JsonArray.clear() {
	while (size() > 0) {
		remove(size() - 1)
	}
}
