package ru.dbotthepony.kstarbound.util

import com.google.gson.JsonObject
import com.google.gson.JsonPrimitive
import com.google.gson.TypeAdapter
import com.google.gson.internal.bind.TypeAdapters
import com.google.gson.stream.JsonReader
import com.google.gson.stream.JsonToken
import com.google.gson.stream.JsonWriter
import ru.dbotthepony.kstarbound.RegistryObject
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.defs.item.api.IItemDefinition
import ru.dbotthepony.kstarbound.io.json.consumeNull

class ItemStack private constructor(item: RegistryObject<IItemDefinition>?, count: Long, val parameters: JsonObject, marker: Unit) {
	constructor(item: RegistryObject<IItemDefinition>, count: Long = 1L, parameters: JsonObject = JsonObject()) : this(item, count, parameters, Unit)

	var item: RegistryObject<IItemDefinition>? = item
		private set

	var size = count
		set(value) {
			if (field == 0L || item == null)
				return

			field = value.coerceAtLeast(0L)

			if (field == 0L) {
				item = null
			}
		}

	val isEmpty: Boolean
		get() = size <= 0 || item == null

	val isNotEmpty: Boolean
		get() = size > 0 && item != null

	val maxStackSize: Long
		get() = item?.value?.maxStack ?: 0L

	fun grow(amount: Long) {
		size += amount
	}

	fun shrink(amount: Long) {
		size -= amount
	}

	/**
	 * Возвращает null если этот предмет пуст
	 */
	fun conciseToNull(): ItemStack? {
		if (isEmpty) {
			return null
		} else {
			return this
		}
	}

	fun mergeFrom(other: ItemStack, simulate: Boolean) {
		if (isStackable(other)) {
			val newCount = (size + other.size).coerceAtMost(item!!.value.maxStack)
			val diff = newCount - size
			other.size -= diff

			if (!simulate)
				size = newCount
		}
	}

	fun lenientEquals(other: Any?): Boolean {
		if (other !is ItemStack)
			return false

		if (isEmpty)
			return other.isEmpty

		return other.size == size && other.item == item
	}

	fun isStackable(other: ItemStack): Boolean {
		return size != 0L && other.size != 0L && item!!.value.maxStack < size && other.item == item && other.parameters == parameters
	}

	override fun equals(other: Any?): Boolean {
		if (other !is ItemStack)
			return false

		if (isEmpty)
			return other.isEmpty

		return other.size == size && other.item == item && other.parameters == parameters
	}

	// мы не можем делать хеш из count и parameters так как они изменяемы
	override fun hashCode(): Int {
		return item.hashCode()
	}

	override fun toString(): String {
		if (isEmpty)
			return "ItemDescriptor.EMPTY"

		return "ItemDescriptor[${item!!.value.itemName}, count = $size, params = $parameters]"
	}

	fun copy(): ItemStack {
		if (isEmpty)
			return this

		return ItemStack(item, size, parameters.deepCopy(), Unit)
	}

	fun toJson(): JsonObject? {
		if (isEmpty) {
			return null
		}

		return JsonObject().also {
			it.add("name", JsonPrimitive(item!!.value.itemName))
			it.add("count", JsonPrimitive(size))
			it.add("parameters", parameters.deepCopy())
		}
	}

	class Adapter(val starbound: Starbound) : TypeAdapter<ItemStack>() {
		override fun write(out: JsonWriter, value: ItemStack?) {
			val json = value?.toJson()

			if (json == null)
				out.nullValue()
			else
				TypeAdapters.JSON_ELEMENT.write(out, json)
		}

		override fun read(`in`: JsonReader): ItemStack {
			if (`in`.consumeNull())
				return EMPTY

			return starbound.item(TypeAdapters.JSON_ELEMENT.read(`in`))
		}
	}

	companion object {
		@JvmField
		val EMPTY = ItemStack(null, 0L, JsonObject(), Unit)
	}
}
