package ru.dbotthepony.kstarbound.world.entities

import ru.dbotthepony.kstarbound.world.Chunk
import ru.dbotthepony.kstarbound.world.World
import ru.dbotthepony.kvector.vector.Vector2d

abstract class Entity(val world: World<*, *>) {
	/**
	 * The chunk this entity currently in, it is automatically updated on each change of [position]
	 */
	var chunk: Chunk<*, *>? = null
		set(value) {
			if (!isSpawned) {
				throw IllegalStateException("Trying to set chunk this entity belong to before spawning in world")
			} else if (isRemoved) {
				throw IllegalStateException("This entity was removed")
			} else if (value == field) {
				return
			}

			val chunkPos = world.chunkFromCell(position)

			if (value != null && chunkPos != value.pos) {
				throw IllegalStateException("Set proper position before setting chunk this Entity belongs to (expected chunk $chunkPos, got chunk ${value.pos})")
			}

			val oldChunk = field
			field = value

			if (oldChunk == null && value != null) {
				world.orphanedEntities.remove(this)
				value.addEntity(this)
			} else if (oldChunk != null && value == null) {
				world.orphanedEntities.add(this)
				oldChunk.removeEntity(this)
			} else if (oldChunk != null && value != null) {
				value.transferEntity(this, oldChunk)
			}
		}


	/**
	 * Current entity position. If entity is logical, this can only be changed
	 * by external means, otherwise it always stands where it is.
	 *
	 * If entity is physical, then movement controller will update this on each physics step
	 * to match position of physical body.
	 *
	 * Setting this value will update [chunk] immediately, if [isSpawned] is true and [isRemoved] is false.
	 */
	var position = Vector2d()
		set(value) {
			if (field == value)
				return

			val old = field
			field = Vector2d(world.x.cell(value.x), world.y.cell(value.y))

			movement.notifyPositionChanged()

			if (isSpawned && !isRemoved) {
				val oldChunkPos = world.chunkFromCell(old)
				val newChunkPos = world.chunkFromCell(field)

				if (oldChunkPos != newChunkPos) {
					chunk = world.chunkMap[newChunkPos]
				}
			}
		}


	/**
	 * This entity's angle in radians.
	 *
	 * Logical entities never rotate.
	 *
	 * Alive entities usually don't rotate.
	 *
	 * If entity is physical, this value is updated by movement controller on
	 * each physics step to match angle of physical body.
	 */
	var angle: Double = 0.0
		set(value) {
			if (field == value)
				return

			field = value
			movement.notifyPositionChanged()
		}


	/**
	 * Whenever is this entity spawned in world ([spawn] called).
	 * Doesn't mean entity still exists in world, check it with [isRemoved]
	 */
	var isSpawned = false
		private set

	/**
	 * Whenever is this entity was removed from world ([remove] called).
	 */
	var isRemoved = false
		private set

	open fun spawn() {
		if (isSpawned)
			throw IllegalStateException("Already spawned")

		isSpawned = true
		world.entities.add(this)
		chunk = world.chunkMap[world.chunkFromCell(position)]

		if (chunk == null) {
			world.orphanedEntities.add(this)
		}

		movement.onSpawnedInWorld()
	}

	open fun remove() {
		if (isRemoved)
			throw IllegalStateException("Already removed")

		isRemoved = true

		if (isSpawned) {
			world.entities.remove(this)
			chunk?.removeEntity(this)
		}

		movement.destroy()
	}

	/**
	 * This entity's movement controller. Even logical entities have one, but they have
	 * dummy movement controller, which does nothing.
	 *
	 * If entity is physical, this controller handle interaction with Box2D world, update angles and
	 * position and other stuff.
	 */
	abstract val movement: MovementController<*>

	/**
	 * Внутренний блок "раздумья" сущности, вызывается на каждом тике мира
	 */
	protected abstract fun thinkAI()

	/**
	 * Заставляет сущность "думать".
	 */
	fun think() {
		if (!isSpawned) {
			throw IllegalStateException("Tried to think before spawning in world")
		}

		movement.think()
		thinkAI()
	}

	open fun onTouchSurface(velocity: Vector2d, normal: Vector2d) {

	}
}
